--[[
YAGRA - gear rating addon
Copyright 2010 Adirelle (adirelle@tagada-team.net)
All rights reserved.
--]]

local addonName, addon = ...
local L = addon.L

local mod = addon:NewModule('ChatFilter', 'AceTimer-3.0', 'AceHook-3.0')
mod.enabledInCombat = true

function mod:OnInitialize()
	self:RegisterDatabase{
		profile = {
			channels = { ['*'] = true },
			appendLevel = false,
		}
	}
end

function mod:GetOptions()
	return {
		_nameChatScore = {
			name = L['Scores in chat messages'],
			type = 'header',
			order = 400,
		},
		appendScore = self:CreateEnableOption{
			name = L['Append score to names'],
			desc = L['Check to append gear score to character names in chat messages.'],
			order = 401,
		},
		appendLevel = {
			name = L['Append level to names'],
			desc = L['Check to append level to character names in chat messages.'],
			type = 'toggle',
			order = 402,
		},
		channels = {
			name = L['Modified channels'],
			desc = L['Select channels to enhance with gear score. Note that disabling guild, party and raid setting applies to resp. guild, party and raid members, wherever they speak.'],
			type = 'multiselect',
			order = 410,
			values = {
				CHANNEL = CHANNEL,
				PARTY = PARTY,
				RAID = RAID,
				GUILD = GUILD,
				SYSTEM = L['System messages'],
				WHISPER = WHISPER,
			},
		},
	}, self:GetOptionHandler()
end

function mod:OnEnable()
	self:RawHook('GetColoredName', true)
	self:ScheduleRepeatingTimer('CleanupCache', 300)
	self:RegisterEvent('PLAYER_UPDATE_RESTING', "CleanupCache")
	self:RegisterEvent('PLAYER_LEAVING_WORLD', "CleanupCache")
	self:RegisterEvent('PLAYER_REGEN_DISABLED', "CleanupCache")
	self:RegisterMessage('YAGRA_RecordUpdated')
end

local playerName = addon.playerName
local function FilterName(fullName)
	if fullName == playerName or
			not mod.db.profile.channels.GUILD and IsInGuild() and UnitIsInMyGuild(fullName) or
			not mod.db.profile.channels.RAID and UnitInRaid(fullName) or
			not mod.db.profile.channels.PARTY and UnitInParty(fullName) then
		return false
	end
	return true
end

local function FormatName(name, record)
	if record then
		local score = addon:FormatScore(record, true, true, true)
		if mod.db.profile.appendLevel then
			return strjoin(':', name, tostring(record.level), score)
		else
			return strjoin(':', name, score)
		end
	else
		return name
	end
end

local counter = 0
local nameCache = setmetatable({}, {
	__index = function(self, name)
		if not name or name == "" then return name end
		local formattedName = name
		if FilterName(name) then
			formattedName = FormatName(name, addon:QueryRecord(name))
		end
		rawset(self, name, formattedName)
		counter = counter + 1
		return formattedName
	end,
})

function mod:CleanupCache()
	if counter > 100 then
		wipe(nameCache)
		counter = 0
	end
end

function mod:YAGRA_RecordUpdated(event, record, source)
	local fullName = record.fullName
	nameCache[fullName] = FilterName(fullName) and FormatName(fullName, record) or fullName
	count = counter + 1
end

function mod:GetColoredName(event, arg1, name, ...)
	return self.hooks.GetColoredName(event, arg1, name and nameCache[name], ...)
end

