--[[
YAGRA - gear rating addon
Copyright 2010 Adirelle (adirelle@tagada-team.net)
All rights reserved.
--]]

local addonName, addon = ...
local L = addon.L

local UnitNameRealm = addon.UnitNameRealm

local mod = addon:NewModule('Tooltip', 'AceHook-3.0')
mod.enabledInCombat = true

function mod:OnInitialize()
	self:RegisterDatabase{
		profile = {
			showInventoryDetail = 'always',
			altItemScores = true,
		}
	}
	if self.db.profile.unitTooltipDetail then
	end
end

local tooltips = {
	"GameTooltip", "ShoppingTooltip1", "ShoppingTooltip2", "ShoppingTooltip3",
	"SmallTextTooltip", "ItemRefTooltip",
}

function mod:OnEnable()
	for i, name in pairs(tooltips) do
		local tooltip = _G[name]
		if type(tooltip) == "table" and tooltip:IsObjectType('GameTooltip') then
			self:SecureHookScript(tooltip, "OnTooltipSetUnit")
			self:SecureHookScript(tooltip, "OnTooltipSetItem")
		else
			tooltips[i] = nil
		end
	end
	self:RegisterMessage('YAGRA_RecordUpdated')
	self:RegisterEvent('MODIFIER_STATE_CHANGED')
end

function mod:GetOptions()
	return {
		_displayHeader = {
			name = L['Tooltip'],
			type = 'header',
			order = 300,
		},
		showInventoryDetail = {
			name = L['Detailed player tooltip'],
			desc = L['Select when to display all the items of a player, along with their score, in his/her tooltip.'],
			type = 'select',
			values = {
				never = L['Never'],
				always = L['Always'],
				shift = L['When the shift key is pressed'],
				alt = L['When the alt key is pressed'],
				control = L['When the control key is pressed'],
			},
			order = 320,
		},
		altItemScores = {
			name = L['Alternative item scores'],
			desc = L["Display alternative scores (hunter and Titan's Grip) in addition to basic score in item tooltip."],
			type = 'toggle',
			order = 340,
		},
	}, self:GetOptionHandler()
end

local function ShouldShowInventoryDetail()
	local showDetail = mod.db.profile.showInventoryDetail
	return showDetail == 'always' or
		(showDetail == 'shift' and IsShiftKeyDown()) or
		(showDetail == 'alt' and IsAltKeyDown()) or
		(showDetail == 'control' and IsControlKeyDown())
end

function mod:EnhanceUnitTooltip(tooltip, record)
	local score, r, g, b = addon:FormatScore(record)
	tooltip:AddLine(L['Gear score: ']..score, r, g, b)
	if ShouldShowInventoryDetail() then
		for slot = 1, 18 do
			local item = record[slot] or "0:0"
			local name, link, quality, level, _, _, _, _, _, icon = GetItemInfo('item:'..item)
			if link then
				local score = floor(addon:GetPlayerItemScore(record, item, slot, addon.db.profile.alternativeRating)+0.5)
				local color = quality and ITEM_QUALITY_COLORS[quality]
				local hexcolor = color and color.hex or ""
				tooltip:AddDoubleLine(format("|T%s:16:16:0:0:64:64:4:60:4:60|t %s%s (%d)|r", icon, hexcolor, name, level), score)
			end
		end
	end
	tooltip:Show()
end

function mod:MODIFIER_STATE_CHANGED(event, key)
	local showDetail = mod.db.profile.showInventoryDetail
	if (showDetail == 'shift' and (key == 'LSHIFT' or key == 'RSHIFT')) or
			(showDetail == 'alt' and (key == 'LALT' or key == 'RALT')) or
			(showDetail == 'control' and (key == 'LCTRL' or key == 'RCTRL')) then
		for i, name in pairs(tooltips) do
			local tooltip = _G[name]
			if tooltip:IsVisible() then
				local _, unit = tooltip:GetUnit()
				if unit then
					tooltip:SetUnit(unit)
				end
			end
		end
	end
end

function mod:OnTooltipSetUnit(tooltip)
	local _, unit = tooltip:GetUnit()
	if not unit then return end
	local name, realm, fullName = UnitNameRealm(unit)
	local valid, reason = addon:IsValidUnit(unit)
	if not valid then
		return self:Debug('Ignoring tooltip', fullName, unit, reason)
	end
	local record = addon:QueryRecord(name, realm, unit)
	if record then
		self:EnhanceUnitTooltip(tooltip, record)
	end
end

function mod:OnTooltipSetItem(tooltip)
	local name, link = tooltip:GetItem()
	if not link then return end
	local equipSlot = select(9, GetItemInfo(link))
	if not equipSlot or equipSlot == "" or equipSlot == "INVTYPE_BAG" or equipSlot == "INVTYPE_TABARD" then return end
	local item = link:match('item:(%d+:%d+)')
	local score = item and addon:GetItemScore(item)
	if not score then return end
	score = math.floor(score+0.5)
	tooltip:AddLine('Item score: '..score)
	if self.db.profile.altItemScores then
		local hunterScore = math.floor(addon:GetItemScore(item, true)+0.5)
		if hunterScore ~= score then
			tooltip:AddLine(L['Hunter item score: %d']:format(hunterScore))
		end
		local tgScore = math.floor(addon:GetItemScore(item, false, true)+0.5)
		if tgScore ~= score then
			tooltip:AddLine(L["Titan's Grip item score: %d"]:format(tgScore))
		end
	end
end

function mod:YAGRA_RecordUpdated(event, record, source)
	for i, name in pairs(tooltips) do
		local tooltip = _G[name]
		if tooltip:IsVisible() then
			local _, unit = tooltip:GetUnit()
			if unit then
				local name, realm = UnitNameRealm(unit)
				if name == record.name and realm == record.realm then
					tooltip:SetUnit(unit)
				end
			end
		end
	end
end

